#include "parser.h"
#include "printer.h"
#include "polynomial.h"
#include "division.h"
#include "buchberger.h"
#include "wallideal.h"
#include "lp.h"
#include "reversesearch.h"
#include "termorder.h"
#include "ep_standard.h"
#include "ep_xfig.h"
#include "gfanapplication.h"

class FacetsApplication : public GFanApplication
{
  //  SimpleOption optionTestCdd;
public:
  const char *helpText()
  {
    return "This program computes the facet normals of a full-dimensional cone specified on the input. The input is a list of vectors each defining a half space. The cone is the intersection of these half spaces. The output is a minimal list of normals defining the cone.\n";
  }
  FacetsApplication()
    //:    optionTestCdd("-cdd","undocumented")
  {
    registerOptions();
  }
    
  char *name()
  {
    return "_facets";
  }
  int main()
  {
    LpSolver::printList(Stdout);
    lpSetSolver("cddgmp");
    

    FileParser P(Stdin);

    IntegerVectorList normals=P.parseIntegerVectorList();

    /*    if(optionTestCdd.getValue())
      {
	AsciiPrinter(Stdout).printVectorList(extremeRaysInequalityIndices(normals));
	
	return 0;
      }
    */
    normals=wallRemoveScaledInequalities(normals);
    IntegerVectorList facets;
    for(IntegerVectorList::const_iterator i=normals.begin();i!=normals.end();i++)
      //if(wallContainsPositiveVector(*i))
      if(isFacet(normals,i))
        {
	  facets.push_back(*i);
        }
    
    AsciiPrinter(Stderr).printVectorList(facets);
 
    fprintf(Stderr,"Rank:%i\n",rankOfMatrix(facets));

    return 0;
  }
};

static FacetsApplication theApplication;
