#include "parser.h"
#include "printer.h"
#include "polynomial.h"
#include "division.h"
#include "buchberger.h"
#include "wallideal.h"
#include "lp.h"
#include "reversesearch.h"
#include "polyhedralfan.h"
#include "breadthfirstsearch.h"
#include "termorder.h"
#include "ep_standard.h"
#include "ep_xfig.h"
#include "gfanapplication.h"
#include "timer.h"

class FVectorApplication : public GFanApplication
{
  FieldOption theFieldOption;
  IntegerOption debugOption;
public:
  const char *helpText()
  {
    return "This program takes the complete list of reduced Groebner bases for a homogeneous ideal and outputs the f-vector for the Groebner fan.\n";
  }
  FVectorApplication():
    debugOption("-d","Tell the program to print out the partial f-vector for every nth Groebner basis. Besides printing the partial f-vector on the stderr the program also tries to write it to the file gfan_fvector_temporary.",0)
  {
    registerOptions();
  }    

  char *name()
  {
    return "_fvector";
  }

  int main()
  {
    Field::printList(Stderr);

    LpSolver::printList(Stderr);
    lpSetSolver("cddgmp");
     
    FileParser p(Stdin);

    int c=p.nextNonBlank();
    
    IntegerVector f(0);
    int homog=-1;
    int n=-1;

    int counter=0;
    assert(p.isLeftBrackets(c));
    do
      {
	PolynomialSet g=p.parsePolynomialSet();
	if(f.size()==0)
	  {
	    n=g.numberOfVariablesInRing();
	    homog=dimensionOfHomogeneitySpace(g);
	    f=IntegerVector(n-homog+1);
	    //	    fprintf(stderr,"Dimensions computed.\n");
	  }
	PolyhedralCone c=groebnerCone(g);
	PolyhedralFan F(n);
	F.insert(c);
	for(int i=0;i<n-homog+1;i++)
	  {
	    F.removeAllLowerDimensional();
	    //	    IntegerVectorList l=F.getRelativeInteriorPoints();
	    //	    if(i==n-homog)
	    //  {
	    //	assert(l.size()==1);
	    //  }
	    PolyhedralFan F2(n);
	    while(!F.isEmpty())
	      {
		//		fprintf(stderr,"checking cone\n");
		PolyhedralCone K=F.highestDimensionalCone();
		F.remove(K);

		WeightTermOrder myOrder(K.getRelativeInteriorPoint());
		if(g.checkMarkings(myOrder))
		  {
		    f[n-homog-i]++;
		    F2.insert(K);
		  }
	      }
	    F=F2.facetComplex();
	    //	    fprintf(stderr,"facetComplex computed\n");
	  }
	counter++;
	if(debugOption.getValue())
	  {
	    if((counter%debugOption.getValue())==0)
	      {
		AsciiPrinter(Stderr).printVector(f);
		AsciiPrinter(Stderr).printNewLine();

		FILE *p=fopen("gfan_fvector_temporary","w");
		if(p)
		  {
		    AsciiPrinter(p).printVector(f);
		    AsciiPrinter(p).printNewLine();
		    fclose(p);
		  }
	      }
	  }
      }
    while((c=p.nextNonBlank())==',');
    assert(p.isRightBrackets(c));

    AsciiPrinter(Stdout).printVector(f);
    return 0;
  }
};

static FVectorApplication theApplication;

